require 'active_model'

module QueryInterface
  module Client
    class Property
      attr_accessor :type, :name, :primary, :update, :array, :default

      def initialize(name, type, options={})
        unless conversion = QueryInterface::Client.types[type]
          raise "Unknown Type #{type}"
        end
        self.type = conversion
        self.primary = options.delete(:primary) { false }
        self.update = options.delete(:update) { true }
        self.array = options.delete(:array) { false }
        self.default = options.delete(:default) { lambda { nil } }
      end

      def run_get_conversion(value)
        if self.array
          value ||= []
          value.map { |item| self.type.apply_get(item) }
        else
          self.type.apply_get(value) unless value.nil?
        end
      end

      def run_set_conversion(value)
        if self.array
          value ||= []
          value.map { |item| self.type.apply_set(item) unless item.blank? }
            .compact
        else
          self.type.apply_set(value) unless value.nil?
        end
      end
    end

    module Model
      def self.included(base)
        base.include(QueryInterface::Client::Request)
        base.include(QueryInterface::Client::Resource)
        base.include(ActiveModel::Validations)
        base.include(ActiveModel::Validations::Callbacks)
        base.include(ActiveModel::Conversion)
        base.instance_variable_set(:@properties, {})
        base.send(:attr_accessor, :_property_values, :_additional_values, :_original_values, :_destroyed)
        base.extend(ClassMethods)
        base.include(InstanceMethods)
      end

      module ClassMethods
        def property(name, type, options={}, &definitions)
          property = QueryInterface::Client::Property.new(name, type, options)

          define_method name do
            self.get_value(name)
          end

          unless property.primary || !property.update
            define_method "#{name}=".to_sym do |value|
              self.set_value(name, value)
            end
          end

          if block_given?
            property.instance_eval &definitions
          end

          @properties[name] = property
        end

        def primary_keys
          @properties.map do |name, property|
            name if property.primary
          end.compact
        end

        def new_collection(collection)
          collection[:data].map do |data|
            self.new_from_hash(data)
          end
        end

        def get(path, params={})
          data = self.get_raw(path: path, params: params)[:parsed_data][:data]
          self.new_from_hash(data)
        end

        def new_from_hash(data={})
          instance = self.new
          instance.assign_attributes(data)
          instance._original_values = instance._property_values.deep_dup
          instance
        end
      end

      module InstanceMethods
        def initialize(data={})
          self.properties.each do |name, property|
            default = self.instance_exec(&property.default)
            self.set_value(name, default)
          end
          self.assign_attributes(data, run_conversions: false)
          self._original_values = self._property_values.deep_dup
        end

        def properties
          self.class.instance_variable_get(:@properties)
        end

        def assign_attributes(data, options={})
          run_conversions = options.delete(:run_conversions) {true}
          data.each do |name, value|
            name = name.to_sym
            if self.properties.has_key?(name)
              property = self.properties[name]
              if run_conversions
                begin
                  converted = property.run_set_conversion(value)
                rescue
                  self.set_value(name, value)
                  self.errors.add(name, "Type conversion failed")
                else
                  self.set_value(name, converted)
                end
              else
                self.set_value(name, value)
              end
            else
              self.additional_values[name] = value
            end
          end
        end

        def to_attributes(additional_data={})
          data = {}
          properties = self.properties.map do |name, property|
            [name, property.run_get_conversion(self.get_value(name))] if (property.update && !property.primary)
          end.compact.to_h
          data[self.class.model_name.param_key.to_sym] = properties.merge(additional_data)
          data
        end

        def save(additional_data={})
          return false unless (self.errors.empty? && self.valid?)
          return true if self.persisted? && !self.dirty?

          if self.persisted?
            result = self.put_raw(params: self.to_attributes(additional_data))
          else
            result = self.class.post_raw(params: self.to_attributes(additional_data))
            if result[:response].success?
              self.assign_attributes(result[:parsed_data][:data])
            end
          end
          assign_errors(result)
          status = result[:response].success?
          self._original_values = self._property_values.deep_dup if status

          status
        end

        def update_attributes(attributes)
          self.assign_attributes(attributes)
          self.save
        end

        def property_values
          self._property_values ||= {}
        end

        def additional_values
          self._additional_values ||= {}
        end

        def destroyed?
          self._destroyed ||= false
        end

        def property_changes
          self.properties.map do |name, property|
            if self.get_value(name) != self._original_values[name]
              [name, {original: self._original_values[name], new: self.get_value(name)}]
            else
              next
            end
          end.compact.to_h
        end

        def dirty?
          self.property_changes.size > 0
        end

        def set_value(name, value)
          self.errors.delete(name)
          self.property_values[name] = value
        end

        def get_value(name)
          self.property_values[name]
        end

        def persisted?
          !!self.id
        end

        def as_json(options={})
          prop_vals = self.properties.map do |name, property|
            [name, property.run_get_conversion(self.get_value(name))]
          end.to_h
          prop_vals.merge(self.additional_values).as_json(options)
        end

        def get(path, params={})
          self.get_raw(path: path, params: params)[:parsed_data]
        end

        def destroy
          result = self.delete_raw()
          assign_errors(result)
          self._destroyed = !self.errors.any?
        end

      private

        def assign_errors(result)
          (result[:parsed_data][:errors] || {}).each do |name, values|
            values.each { |value| self.errors.add(name, value) }
          end
        end

      end
    end
  end
end
